/*
*   wav_to_tap2.c
*/
#define VERSION "13.2.2004"
/*
*   convert Commodore plus/4 (PAL) tapes in RIFF WAV 8 bit mono format
*   to TAP version 2 format for emulators with maximum accuracy
*/
#define CPR "(c) 2004 Dr.-Ing. Ingo D. Rullhusen <d01c@uni-bremen.de>"
/*
* This program is free software; you can redistribute it and/or modify   
* it under the terms of the GNU General Public License as published by   
* the Free Software Foundation; either version 2 of the License, or   
* (at your option) any later version.   
* 
* This program is distributed in the hope that it will be useful,   
* but WITHOUT ANY WARRANTY; without even the implied warranty of   
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   
* GNU General Public License for more details.   
*
* You should have received a copy of the GNU General Public License   
* along with this program; if not, write to the Free Software   
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*
*/


#include <stdlib.h>
#include <stdio.h>


/* taken from Markus Brenner's mtap */

/* The following frequencies in [Hz] determine the length of one TAP unit */
#define C64PALFREQ  123156      /*  985248 / 8 */
#define C64NTSCFREQ 127841      /* 1022727 / 8 */
#define VICPALFREQ  138551      /* 1108405 / 8 */
#define VICNTSCFREQ 127841      /* 1022727 / 8 */
#define C16PALFREQ  110840      /*  886724 / 8 */
#define C16NTSCFREQ 111860      /*  894886 / 8 */ 

/* Machine entry used in extended TAP header */   
#define C64 0
#define VIC 1
#define C16 2

/* Video standards */
#define PAL  0
#define NTSC 1

/* end of taken */


#define MAXTH 0xc0  /* according to Attila Grosz */
#define MINTH 0x3f  /* dito                      */


struct WAV_HEADER
{
   char riff[4];
   unsigned int  f_length_m8;
   char wave[4];
   char fmt_[4];
   unsigned int  fmt_length;
   unsigned short format;
   unsigned short channels;
   unsigned int sample_rate;
   unsigned int byte_per_sec;
   unsigned short block_align;
   unsigned short bits_per_sample;
   char data[4];
   unsigned int data_length;
} wav_header;


struct TAP_HEADER
{
   char identifier[12];
   unsigned char tapvers;
   unsigned char machine;
   unsigned char video_standard;
   unsigned char dummy;
   unsigned long filelength;
} tap_header = { { 'C','1','6','-','T','A','P','E','-','R','A','W' },
                 2,
                 C16,
                 PAL,
                 0,
                 0
               };


int write_pulse( unsigned long long longpulse, FILE *tap )
{
   int i, count;
   unsigned long long zerot;

   count = 0;

   zerot = longpulse >> 3;

   if ( zerot < 256 && zerot > 0 )
   {
      fputc(zerot, tap);
      longpulse = 0;
      count++;
   }

   while ( longpulse != 0 )
   {
      if ( longpulse >= 0x1000000 )
         zerot = 0xffffff;    /* zerot = maximum */
      else
         zerot = longpulse;	

      longpulse -= zerot;

      fputc(0, tap);
      for ( i=0; i<3; i++ )
      {
         fputc(zerot % 256, tap);
         zerot /= 256;
      }
      count++;
   }

   return(count);
}


int main( int argc, char *argv[] )
{
   char *wavfile, *tapfile;

   double relshortpos;

   short old_state, state, ready;

   unsigned long long sample_rate, t_rate,
                      tap_pos, wav_pos, pulse_start, lpulse, longpos;

   unsigned char byte, byte0, byte1;

   FILE *WAV, *TAP;


   if ( argc != 3 )
   {
      printf("wav_to_tap2 (version %s)\n",VERSION);
      printf(" Copyright %s\n",CPR);
      printf(" wav_to_tap2 comes with ABSOLUTELY NO WARRANTY; for details see GNU\n");
      printf(" GENERAL PUBLIC LICENSE. This is free software, and you are welcome to\n");
      printf(" redistribute it under certain conditions; see GNU GENERAL PUBLIC LICENSE\n");
      printf(" for details.\n\n");
      printf("wav_to_tap2 <WAV file> <TAP file>\n");
      exit(1);
   }
   printf("wav_to_tap2 (version %s)\n",VERSION);
   printf(" Copyright %s\n",CPR);
   printf(" wav_to_tap2 comes with ABSOLUTELY NO WARRANTY; for details see GNU\n");
   printf(" GENERAL PUBLIC LICENSE. This is free software, and you are welcome to\n");
   printf(" redistribute it under certain conditions; see GNU GENERAL PUBLIC LICENSE\n");
   printf(" for details.\n\n");

   wavfile = argv[1];
   tapfile = argv[2];

   WAV = fopen(wavfile,"r");
   if ( WAV==NULL )
   {
      printf("cannot open %s\n",wavfile);
      exit(1);
   }
   TAP = fopen(tapfile,"w");
   if ( TAP==NULL )
   {
      printf("cannot open %s\n",tapfile);
      exit(1);
   }
                        

   fread((void*)&wav_header,sizeof(wav_header),1,WAV);

   if ( wav_header.riff[0] != 'R'
        || wav_header.riff[1] != 'I'
           || wav_header.riff[2] != 'F'
              || wav_header.riff[3] != 'F' ||
        wav_header.wave[0] != 'W'
        || wav_header.wave[1] != 'A'
           || wav_header.wave[2] != 'V'
              || wav_header.wave[3] != 'E' ||
        wav_header.fmt_[0] != 'f'
        || wav_header.fmt_[1] != 'm'
           || wav_header.fmt_[2] != 't'
              || wav_header.fmt_[3] != ' ' ||
        wav_header.fmt_length != 16 ||
        wav_header.format != 1 ||
        wav_header.channels != 1 ||
        wav_header.bits_per_sample != 8 ||
        wav_header.data[0] != 'd' 
        || wav_header.data[1] != 'a'
           || wav_header.data[2] != 't'
              || wav_header.data[3] != 'a'
      )
   {
      printf("Error: input not a simple RIFF WAVE PCM 8 bit mono!\n");
      exit(1);
   }

   sample_rate = (unsigned long long)( wav_header.sample_rate );

   printf("input sample rate: %llu Hz\n",sample_rate);

   fwrite((void*)&tap_header,sizeof(tap_header),1,TAP);

   t_rate = 8 * (unsigned long long)C16PALFREQ;

   ready = 0;
   old_state = state = 0;
   tap_pos = 0;

   fread( (void*)&byte0, sizeof(byte), 1, WAV );
   fread( (void*)&byte1, sizeof(byte), 1, WAV );
   wav_pos = 1;

   longpos = 0;
   while ( !feof(WAV) )
   {
      if ( longpos*sample_rate > wav_pos*t_rate )
      {
         byte0 = byte1;
         fread( (void*)&byte1, sizeof(byte), 1, WAV );
         wav_pos++;
      }

      relshortpos = (double)( longpos*sample_rate
                              - wav_pos*t_rate + t_rate ) / (double)t_rate;

      byte = (unsigned char)( byte1*relshortpos
                              + byte0*(1.0-relshortpos) );

      if ( byte >= MAXTH )
      {
         old_state = state;
         state = +1;
      }
      else if ( byte <= MINTH )
      {
         old_state = state;
         state = -1;
      }
      else
      {
         old_state = state;
      }

      if ( state > old_state )
      {
         if ( ready == 0 )
         {
            pulse_start = longpos;
            ready = 1;
         }
         else
         {
            lpulse = longpos - pulse_start;
            tap_pos += write_pulse(lpulse, TAP);
            pulse_start = longpos;
         }
      }

      if ( state < old_state && ready == 1 )
      {
            lpulse = longpos - pulse_start;
            tap_pos += write_pulse(lpulse, TAP);
            pulse_start = longpos;
      }

      longpos++;
   }

   printf("written halfpulses: %llu\n",tap_pos);

   tap_header.filelength = ftell(TAP) - 20;
   fseek(TAP, 0, SEEK_SET);
   fwrite((void*)&tap_header,sizeof(tap_header),1,TAP);

   fclose(TAP);

   fclose(WAV);

   return(0);
}
